package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.glass.opengl.enums.GLGetPName;
import dev.pfaff.log4truth.NamedLogger;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.ref.SoftReference;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.LockSupport;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.util.BasicFuture.completedTask;
import static dev.pfaff.altargui.util.ThreadFuture.spawnVirtualTask;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.ERROR;

public final class TextureAllocator implements AutoCloseable {
	private static final NamedLogger LOGGER = new NamedLogger("TexAlloc");

	public static final int MAX_TEXTURE_SIZE;

	static {
		MAX_TEXTURE_SIZE = GL.glGetInteger(GLGetPName.MAX_TEXTURE_SIZE);
		if (MAX_TEXTURE_SIZE == 0) {
			throw new IllegalStateException("MAX_TEXTURE_SIZE must be greater than 0");
		}
	}

	/**
	 * The minimum allocation size in pixels. Herein we often measure in "units" rather than pixels for efficiency.
	 */
//	public static final int UNIT = 8;
	public static final int UNIT = Math.divideExact(MAX_TEXTURE_SIZE, 256);

	static {
		if (Integer.bitCount(UNIT) != 1) {
			throw new UnsupportedOperationException(STR."Only power-of-two unit sizes are supported, but \{UNIT} was calculated");
		}
	}

	/**
	 * The size (in {@link #UNIT}s) of each layer in the allocator.
	 */
	private final int sizeUnits;
	private int layers;
	private boolean ab;
	private final TextureArray textureArrayA = new TextureArray();
	private final TextureArray textureArrayB = new TextureArray();
	private final BitSet allocations = new BitSet();
	private final Object2ObjectOpenHashMap<LoadableTexture, SoftReference<TextureHandle>> cache =
			new Object2ObjectOpenHashMap<>();
	private final Semaphore lock = new Semaphore(1);

	/**
	 * @param size the width and height of each layer.
	 */
	public TextureAllocator(int size, int initialLayers) {
		if (size > MAX_TEXTURE_SIZE) {
			throw new IllegalArgumentException("Layer size too large");
		}
		this.sizeUnits = size / UNIT;
		if (size % UNIT != 0) throw new IllegalArgumentException(STR."Invalid layer size: \{size}");
		this.layers = initialLayers;
	}

	public void reset() {
		this.reset(this.getArray());
	}

	private void reset(TextureArray textureArray) {
		var size = this.sizeUnits * UNIT;
		textureArray.allocate(size, size, this.layers, TextureOptions.DEFAULT);
	}

	private void flip() {
		this.ab = !this.ab;
	}

	private TextureArray getArray() {
		return this.ab ? this.textureArrayB : this.textureArrayA;
	}

	private TextureArray getBackArray() {
		return this.ab ? this.textureArrayB : this.textureArrayA;
	}

	public InfallibleAutoClosable bound() {
		return this.getArray().bound();
	}

	/**
	 * Gets a shader to the given texture. If not loaded, the texture will be loaded and a slot allocated for it.
	 */
	public BasicFuture<TextureHandle> alloc(LoadableTexture key) {
		{
			var ref = this.cache.get(key);
			// this inspection is wrong
			//noinspection ConstantValue
			if (ref != null) {
				var handle = ref.get();
				if (handle != null) return completedTask(handle);
			}
		}
		return spawnVirtualTask(() -> {
			ImageData image;
			try {
				image = key.load();
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture load failed", t, List.of(ERROR));
				return Objects.requireNonNull(this.cache.get(SinglePixelLoadableTexture.WHITE).get());
			}
			this.lock.acquire();
			try {
				try {
					// try again to get from cache, but if not, allocate a new image in the texture array.
					var handle = this.cache.compute(key, (k, ref) -> {
						if (ref != null) {
							var h = ref.get();
							if (h != null) return ref;
						}
						return new SoftReference<>(this.allocNew(k, image));
					}).get();
					if (handle != null) {
						return handle;
					}
					throw new OutOfMemoryError(
							"Handle freed between return from remappingFunction and return from compute. The system is " +
							"probably running very low on memory.");
				} catch (Throwable t) {
					LOGGER.log(() -> "Texture allocation failed", t, List.of(ERROR));
					return Objects.requireNonNull(this.cache.get(SinglePixelLoadableTexture.WHITE).get());
				}
			} finally {
				this.lock.release();
			}
		});
	}

	private TextureHandle allocNew(LoadableTexture key, ImageData image) {
		int width = image.info().width();
		int height = image.info().height();

		// align size up to the next multiple of UNIT.
		// these are in units, not pixels
		int unitWidth = (width + (UNIT - 1)) / UNIT;
		int unitHeight = (height + (UNIT - 1)) / UNIT;
		assert unitWidth * UNIT >= width && unitHeight * UNIT >= height : STR."width and/or height was improperly aligned to unit size:\n\t\{width} > \{unitWidth * UNIT}\n\tor\n\t\{height} > \{unitHeight * UNIT}";

		int sizeUnits = this.sizeUnits;

		if (unitWidth > sizeUnits || unitHeight > sizeUnits) {
			throw new IllegalArgumentException(STR."Size of image exceeds size of allocator layers: \{width} || \{height} > \{this.sizeUnits * UNIT}");
		}

		LOGGER.log(() -> STR."Searching for a free region to allocate \{key} an image of \{width}x\{height}", List.of(DEBUG));

		// find a free region
		int allocationsSize = sizeUnits * sizeUnits * this.layers;
		int fromSearchUpr = allocationsSize - unitWidth - (unitHeight - 1) * sizeUnits;
		int layers = this.layers;
		int sizeUnitsSq = sizeUnits * sizeUnits;

		int xUpr = sizeUnits - unitHeight;

		int from = -1, iUpr = -1;

		search: for (int layerStart = 0; layerStart < layers * sizeUnitsSq; layerStart += sizeUnitsSq) {
			from = layerStart;
			int layerUpr = from + sizeUnitsSq;
			while (from <= fromSearchUpr) {
				assert from % sizeUnits == 0 :
						STR."not at the start of a row: from=\{from}";
				int rowUpr = from + xUpr;
				for (; from < rowUpr; from++) {
					iUpr = from + unitHeight * sizeUnits;
					assert iUpr <= layerUpr;

					if (isRegionFree(from, iUpr, unitWidth)) break search;
				}
				from += unitHeight;
			}
			iUpr = -1;
		}

		if (iUpr == -1) {
			throw new UnsupportedOperationException("TODO: realloc");
		}
		assert from != -1;

		var detailed = this.makeHandle(from, iUpr, unitWidth, width, height);

		var handle = detailed.handle();
		LOGGER.log(() -> STR."Allocated texture handle \{handle.layer}:\{handle.u},\{handle.v} for image of \{width}x\{height} at \{detailed.u() * UNIT},\{detailed.v() * UNIT}", List.of(DEBUG));
		var array = this.getArray();
		BasicFuture.executeOn(() -> {
			try (var _ = this.bound()) {
				array.startUploading();
				array.upload(handle.layer,
						detailed.u() * UNIT,
						detailed.v() * UNIT,
						width,
						height,
						image.info().colorType().glPixelFormat(),
						image.info().colorType().glChannelType(),
						image.pixels());
			}
		}, GLThread::submitToGLThread);
		LOGGER.log(() -> STR."Uploaded image to \{handle.layer}:\{handle.u},\{handle.v}", List.of(DEBUG));
		return handle;
	}

	private DetailedHandle makeHandle(int from, int iUpr, int unitWidth, int textureWidth, int textureHeight) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// mark allocated
		for (int i = from; i < iUpr; i += sizeUnits) { // for each row
			// mark the region of each row allocated
			allocations.set(i, i + unitWidth);
		}

		var handle = MemorySegment.NULL.reinterpret(Arena.ofAuto(), _ -> {
			this.lock.acquireUninterruptibly();
			try {
				// mark free
				for (int i = from; i < iUpr; i += sizeUnits) {
					allocations.clear(i, i + unitWidth);
				}
			} finally {
				this.lock.release();
			}
		});

		int layer = from / (sizeUnits * sizeUnits);
		int u = from % sizeUnits;
		int v = (from / sizeUnits) % sizeUnits;
		return new DetailedHandle(handle, layer, sizeUnits, u, v, textureWidth, textureHeight);
	}

	private boolean isRegionFree(int from, int iUpr, int unitWidth) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// check if all are free
		for (int i = from; i < iUpr; i += sizeUnits) {
			for (int x = 0; x < unitWidth; x++) {
				if (allocations.get(i + x)) return false;
			}
		}
		return true;
	}

	@Override
	public void close() {
		this.textureArrayA.close();
		this.textureArrayB.close();
	}

	public void dump() {
		var arr = this.getArray();
//		try (var _ = arr.bound()) {
//			try (var img = new NativeImage(NativeImage.Format.RGBA, this.sizeUnits * UNIT, this.sizeUnits * UNIT * this.layers, false)) {
//				var ms = getMemorySegmentFromImage(img);
//				GL.gl.glGetTexImage(arr.target(), 0, GLPixelFormat.RGBA, GLPixelType.UNSIGNED_BYTE, new MutPointer<>(ms));
//				img.writeTo(Path.of("allocated-texture.png"));
//			} catch (IOException e) {
//				throw new RuntimeException(e);
//			}
//		}
		throw new UnsupportedOperationException("TODO");
	}

	/**
	 * @param u measured in {@link #UNIT}s
	 * @param v measured in {@link #UNIT}s
	 */
	record DetailedHandle(TextureHandle handle, int u, int v) {
		/**
		 * @param layer index
		 * @param layerSizeInUnits the size of each layer, measured in {@link #UNIT}s
		 * @param u measured in {@link #UNIT}s
		 * @param v measured in {@link #UNIT}s
		 */
		public DetailedHandle(MemorySegment handle, int layer, int layerSizeInUnits, int u, int v, int textureWidth, int textureHeight) {
			this(new TextureHandle(handle, layer, layerSizeInUnits * UNIT, u / (float)layerSizeInUnits, v / (float)layerSizeInUnits, textureWidth, textureHeight), u, v);
		}
	}
}
